# copyright ############################### #
# This file is part of the Xdeps Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #
import pickle

import xdeps
from xdeps.tasks import AttrDict


def test_set():
    v = {'a': 1, 'b': 3}
    mgr = xdeps.Manager()
    r = mgr.ref(v)
    r['c'] = 0.1 * r['a'] + 0.2 * r['b']
    assert v['c'] == 0.1 * v['a'] + 0.2 * v['b']
    r['a'] = 1.2
    assert v['c'] == 0.1 * v['a'] + 0.2 * v['b']


def test_attrref():
    class E:
        pass

    e = E()
    e.knl = [0, 1, 2, 3]
    v = {4: e, 2: 1.2}

    mgr = xdeps.Manager()
    v_ = mgr.ref(v, 'v')
    v_[4].knl[1] = v_[2] * 3

    assert v[4].knl[1] == v[2] * 3
    v_[2] = 1.1
    assert v[4].knl[1] == v[2] * 3


def test_inplace():
    m = xdeps.Manager()

    s = {'a': 1, 'b': 2}
    s_ = m.ref(s, 's')
    s_._exec('c=a+b')
    assert s['c'] == s['a'] + s['b']
    s_['c'] += 1  ## s_['a']= s_['c']._expr + 1
    assert s['c'] == s['a'] + s['b'] + 1
    s_['a'] += 1  ## s_['a']= s['a'] + 1
    assert s['c'] == s['a'] + s['b'] + 1


def test_unregister_implicit():
    mgr = xdeps.Manager()
    vd = {}
    v = mgr.ref(vd, 'v')
    v['a'] = 1
    v['b'] = v['a'] * 3

    assert v['b'] in mgr.tasks
    assert v['b'] in mgr.rdeps[v['a']]
    assert v['b'] in mgr.deptasks[v['a']]
    assert v['b'] in mgr.tartasks[v['b']]

    v['b'] = 0

    assert v['b'] not in mgr.tasks
    assert v['b'] not in mgr.rdeps[v['a']]
    assert v['b'] not in mgr.deptasks[v['a']]
    assert v['b'] not in mgr.tartasks[v['b']]


def test_unregister():
    mgr = xdeps.Manager()

    container = {
        'a': 1,
        'b': 2,
        'c': {
            'x': 4,
            'y': 6,
        }
    }

    ref = mgr.ref(container, 'ref')
    ref['c']['x'] = 2 * ref['a'] * ref['b']
    ref['c']['y'] = 3 * ref['a'] * ref['b']

    ref['a'] = 2
    ref['b'] = 2

    assert container['c']['x'] == 8
    assert container['c']['y'] == 12

    mgr.unregister(ref['c']['y'])

    ref['a'] = 1

    assert container['c']['x'] == 4
    assert container['c']['y'] == 12  # unchanged


def test_collisions():
    num_elements = 100000

    mgr = xdeps.Manager()

    elements = {f'bend{ii}': AttrDict(k1=0) for ii in range(num_elements)}
    element_refs = mgr.ref(elements, 'element_refs')

    vars = {}
    var_refs = mgr.ref(vars, 'var_refs')

    for ii in range(num_elements):
        var_refs[f'k{ii}'] = 1
        element_refs[f'bend{ii}'].k0 = var_refs[f'k{ii}']
        assert len(mgr.tasks) == ii + 1

    assert len(mgr.tasks) == num_elements



import numpy as np

from xdeps import Table

data = {
    "name": np.array(["ip1", "ip2", "ip3"]),
    "s": np.array([1.0, 2.0, 3.0]),
    "betx": np.array([4.0, 5.0, 6.0]),
    "bety": np.array([2.0, 3.0, 4.0]),
}


t = Table(data)

def test_column_selection():
    assert len(t.betx)==len(data['betx'])
    assert t["betx", 0] == data['betx'][0]
    assert t["sqrt(betx)"][1]==np.sqrt(data['betx'][1])
    assert t["betx bety"].betx[0] == t.betx[0]
    assert t[["betx", "bety"]].betx[0] == t.betx[0]
    assert t["sqrt(betx)>3 sqrt(bety)"]['sqrt(betx)>3'][0]==(np.sqrt(data['betx'][0])>3)
    assert isinstance(t.cols['betx'], Table)
    assert t.cols["betx", "bety"].betx[0] == t.betx[0]


def test_row_selection():
    assert t[:, 1].betx==data['betx'][1]
    assert t[:, [0, 2]].betx[1]==data['betx'][2]
    assert t[:, t.s > 1].betx[1]==data['betx'][t.s > 1][1]
    assert t.rows[t.s > 1, 1].betx[0]==data['betx'][t.s > 1][0]

def test_row_selection_names():
    assert t[:, "ip1"].betx[0]==data['betx'][0]
    assert t[:, "ip[23]"].betx[0]==data['betx'][1]
    assert t[:, "ip.*##1"].betx[0]==data['betx'][1]
    assert t[:, "notthere"]._nrows==0
    assert t[:, ["ip1", "ip2"]].betx[0]==data['betx'][0]

def test_row_selection_names_with_rows():
    assert t.rows["ip2"].betx[0]==data['betx'][1]
    assert t.rows["ip[23]"].betx[0]==data['betx'][1]
    assert t.rows["ip.*##1"].betx[0]==data['betx'][1]
    assert t.rows["notthere"]._nrows==0
    assert t.rows[["ip1", "ip2"]].betx[1]==data['betx'][1]

def test_row_selection_ranges():
    assert t[:, 1:4:3].betx[0] == data['betx'][1]
    assert t[:, 1.5:2.5:"s"].betx[0] == data['betx'][1]
    assert t[:, "ip1":"ip3"].betx[2] == data['betx'][2]
    assert t[:, "ip.*##1":"ip.*##2"].betx[0] == data['betx'][1]
    assert t[:, "ip2%%-1":"ip2%%+1"].betx[0] == data['betx'][0]
    assert t[:, "ip1":"ip3":"name"].betx[0]  == data['betx'][0]
    assert t[:, None].betx[0]  == data['betx'][0]
    assert t[:, :].betx[0]  == data['betx'][0]

def test_row_selection_ranges_with_rows():
    assert t.rows[1:4:3].betx[0] == data['betx'][1]
    assert t.rows[1.5:2.5:"s"].betx[0] == data['betx'][1]
    assert t.rows["ip1":"ip3"].betx[2] == data['betx'][2]
    assert t.rows["ip.*##1":"ip.*##2"].betx[0] == data['betx'][1]
    assert t.rows["ip2%%-1":"ip2%%+1"].betx[0] == data['betx'][0]
    assert t.rows["ip1":"ip3":"name"].betx[0]  == data['betx'][0]
    assert t.rows[None].betx[0]  == data['betx'][0]
    assert t.rows[:].betx[0]  == data['betx'][0]

def test_numpy_string():
    tab=Table(dict(name=np.array(['a', 'b$b']), val=np.array([1,2])))
    assert tab['val', tab.name[1]]==2




# copyright ############################### #
# This file is part of the Xdeps Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

from collections import defaultdict
from copy import deepcopy
import logging
from typing import Set, Hashable

from .refs import BaseRef, MutableRef, ObjectAttrRef, Ref, RefCount
from .utils import plot_pdot
from .utils import AttrDict
from .sorting import toposort

logger = logging.getLogger(__name__)


def dct_merge(dct1, dct2):
    """Merge two dictionaries, in case of conflict dct2 takes precedence."""
    return {**dct1, **dct2}


def _check_root_owner(t, ref):
    """Check if a task `t`, or any of its parents, has `ref` as owner."""
    if hasattr(t, "_owner"):
        if t._owner is ref:
            return True
        else:
            return _check_root_owner(t._owner, ref)
    else:
        return False


class Task:
    """
    An (abstract) class representing a task.

    A task describes an action that modifies the values of a set of
    target references depending on the values stored in the set of
    references which are dependencies, and potentially its internal state.

    Properties
    ----------
    taskid: BaseRef
        The reference that identifies the task, usually an expression that
        depends on the targets and dependencies and specifies the action
        to be performed.
    targets: Set[BaseRef]
        The set of references that are modified by the task.
    dependencies: Set[BaseRef]
        The set of references that the task depends on.
    """

    taskid: Hashable
    targets: Set[BaseRef]
    dependencies: Set[BaseRef]

    def run(self):
        """Execute the task."""
        raise NotImplemented


class FunctionTask(Task):
    """Task that executes a function when its dependencies change.

    Parameters
    ----------
    taskid: Hashable
        The task identifier.
    action: callable
        The function to be executed.
    targets: Set[BaseRef]
        The set of references that are modified by the task.
    dependencies: Set[BaseRef]
        The set of references that the task depends on.
    """
    action: callable

    def __init__(
            self,
            taskid: Hashable,
            action: callable,
            targets: Set[BaseRef],
            dependencies: Set[BaseRef],
    ):
        self.taskid = taskid
        self.action = action
        self.targets = targets
        self.dependencies = dependencies

    def __repr__(self):
        return f"<Task {self.taskid}:{self.dependencies}=>{self.targets}>"

    def run(self):
        return self.action()


class LinearKnob(Task):
    def __init__(self, taskid, source, weights, targets):
        self.taskid = taskid
        self.source = source
        self.dependencies = {source}
        self.targets = targets
        self.prev_value = source._get_value()
        self.weights = weights

    def run(self):
        value = self.source._get_value()
        delta = value - self.prev_value

        for w, t in zip(self.weights, self.targets):
            # t += ... changes the local variable t in place, we must
            # use _set_value instead.
            t._set_value(t._get_value() + w * delta)

        self.prev_value = value


class ExprTask(Task):
    """
    Task that evaluates an expression `expr` and stores the result in `target`.

    Parameters
    ----------
    target: MutableRef
        The target reference where the result of the expression will be stored.
    expr: BaseRef
        The expression to be evaluated.
    """

    def __init__(self, target: MutableRef, expr: BaseRef):
        self.taskid = target
        self.targets = target._get_dependencies()
        self.dependencies = expr._get_dependencies()
        self.expr = expr

    def __repr__(self):
        return f"{self.taskid} = {self.expr}"

    def run(self):
        value = self.expr._get_value()
        self.taskid._set_value(value)

    def info(self):
        """Print information about the task."""
        print(f"#  {self.taskid}._expr._get_dependencies()")
        for pp in self.expr._get_dependencies():
            print(f"   {pp} = {pp._get_value()}")
        print()


class DepEnv:
    """Proxy for modification of refs and access to the underlying data source.

    A convenience class that can be passed both the original data object and
    a reference to it, so that deferred expressions can be assigned to the ref
    object, but data accesses are performed on the original data source.

    Parameters
    ----------
    data: object
        The original data object.
    ref: Ref
        The reference to the original data object.
    """

    __slots__ = ("_data", "_")

    def __init__(self, data, ref):
        object.__setattr__(self, "_data", data)
        object.__setattr__(self, "_", ref)

    def __getattr__(self, key):
        return getattr(self._data, key)

    def __getitem__(self, key):
        return self._data[key]

    def __setattr__(self, key, value):
        self._[key] = value

    def __setitem__(self, key, value):
        self._[key] = value

    def _eval(self, expr):
        return self._._eval(expr)


class Manager:
    """Value dependency manager.

    A manager registers external Python objects with Xdeps and orchestrates the
    actions performed by Xdeps. It keeps track of the dependencies between
    references and tasks, and executes tasks when their dependencies change.
    The graph of dependencies is stored in the following mappings:

    Properties
    ----------
    tasks: dict
        Maps taskid (in case of expressions that's target ref) to task.
    rdeps: dict
        Maps a ref to the set of all refs that (directly) depend on it.
    rtasks: dict
        Maps a task (identified by a taskid) to the set of all tasks (identified
        by their taskids) whose dependencies are the targets of the given task.
    deptasks: dict
        Maps a ref to all tasks that have ref as a direct dependency.
    tartasks: dict
        Maps ref to all tasks that have ref as direct target.
    containers: dict
        Maps label to the controlled container.
    """

    def __init__(self):
        self.tasks = {}
        self.containers = {}
        self.rdeps = defaultdict(RefCount)
        self.rtasks = defaultdict(RefCount)
        self.deptasks = defaultdict(RefCount)
        self.tartasks = defaultdict(RefCount)
        self._tree_frozen = False

    def ref(self, container=None, label="_"):
        """Return a ref to an instance (or dict) associated to a label.

        Label must be unique.
        """
        if container is None:
            container = AttrDict()
        objref = Ref(container,  label, self)
        assert label not in self.containers
        self.containers[label] = objref
        return objref

    def set_value(self, ref, value):
        """Set a value pointed by a ref and execute all tasks that depends on ref.

        If the value is a Ref, create a new task from the ref.
        """
        logger.info("set_value %s %s", ref, value)
        if ref in self.tasks:
            # TODO: Here we assume that if a ref identifies a task, it must be
            #  it's (only) target. What if that's not true?
            self.unregister(ref)
        if isinstance(value, BaseRef):  # value is an expression
            self.register(ExprTask(ref, value))
            value = value._get_value()  # to be updated
        ref._set_value(value)
        self._run_tasks(self.find_tasks(ref._get_dependencies()))

    def _run_tasks(self, tasks):
        for task in tasks:
            logger.info("Run %s", task)
            task.run()

    def register(self, task):
        """Register a new task identified by taskid"""
        # logger.info("register %s",taskid)
        if self._tree_frozen:
            raise ValueError("Expressions and references cannot be changed "
                             "because the tree is frozen (e.g. because "
                             "a variables cache is active)")
        taskid = task.taskid
        self.tasks[taskid] = task
        for dep in task.dependencies:
            # logger.info("%s have an impact on %s",dep,task.targets)
            self.rdeps[dep].extend(task.targets)
            # logger.info("%s is used by T:%s",dep,taskid)
            self.deptasks[dep].append(taskid)
            for deptask in self.tartasks[dep]:
                # logger.info("%s modifies deps of T:%s",deptask,taskid)
                self.rtasks[deptask].append(taskid)

        for tar in task.targets:
            # logger.info("%s is modified by T:%s",tar,taskid)
            self.tartasks[tar].append(taskid)
            other = self.deptasks[tar]
            for deptask in other:
                # logger.info("T:%s modifies deps of T:%s",taskid,deptask)
                self.rtasks[taskid].append(deptask)

    def unregister(self, taskid):
        """Unregister the task identified by taskid"""
        if self._tree_frozen:
            raise ValueError("Expressions and references cannot be changed "
                             "because the tree is frozen (e.g. because "
                             "a variables cache is active)")
        task = self.tasks[taskid]
        for dep in task.dependencies:
            for target in task.targets:
                if target in self.rdeps[dep]:
                    self.rdeps[dep].remove(target)
            if taskid in self.rtasks[dep]:
                self.rtasks[dep].remove(taskid)
            if taskid in self.deptasks[dep]:
                self.deptasks[dep].remove(taskid)
        for tar in task.targets:
            self.tartasks[tar].remove(taskid)
        if taskid in self.rtasks:
            del self.rtasks[taskid]
        del self.tasks[taskid]

    def freeze_tree(self):
        """Freeze the tree of expressions and references."""
        self._tree_frozen = True

    def unfreeze_tree(self):
        """Unfreeze the tree of expressions and references."""
        self._tree_frozen = False

    def find_deps(self, start_set):
        """Find all refs that depend on ref in `start_set`."""
        assert type(start_set) in (list, tuple, set)
        deps = toposort(self.rdeps, start_set)
        return deps

    def find_taskids_from_tasks(self, start_tasks=None):
        """Find all taskids whose dependencies are affected by the tasks in start_tasks"""
        if start_tasks is None:
            start_tasks = self.rtasks
        tasks = toposort(self.rtasks, start_tasks)
        return tasks

    def find_taskids(self, start_deps=None):
        """Find all taskids that depend on the refs in start_deps"""
        if start_deps is None:
            start_deps = self.rdeps
        start_tasks = set()
        for dep in start_deps:
            start_tasks.update(self.deptasks[dep])
        tasks = toposort(self.rtasks, start_tasks)
        return tasks

    def find_tasks(self, start_deps=None):
        """Find all tasks that depend on the refs in start_deps"""
        if start_deps is None:
            start_deps = self.rdeps
        return [self.tasks[taskid] for taskid in self.find_taskids(start_deps)]

    def iter_expr_tasks_owner(self, ref):
        """Return all ExprTask defintions that write registered container"""
        for t in self.find_tasks():
            # TODO check for all targets or limit to ExprTask
            if _check_root_owner(t.taskid, ref):
                yield str(t.taskid), str(t.expr)

    def copy_expr_from(self, mgr, name, bindings=None):
        """
        Copy expression from another manager

        name: one of toplevel container in mgr
        bindings: dictionary mapping old container names into new container refs
        """
        ref = mgr.containers[name]
        if bindings is None:
            cmbdct = self.containers
        else:
            cmbdct = dct_merge(self.containers, bindings)
        self.load(mgr.iter_expr_tasks_owner(ref), cmbdct)

    def mk_fun(self, name, **kwargs):
        """Write a python function that executes a set of tasks in order of dependencies:
        name: name of the functions
        kwargs:
            the keys are used to defined the argument name of the functions
            the values are the refs that will be set
        """
        varlist, start = list(zip(*kwargs.items()))
        tasks = self.find_tasks(start)
        fdef = [f"def {name}({','.join(varlist)}):"]
        for vname, vref in kwargs.items():
            fdef.append(f"  {vref} = {vname}")
        for tt in tasks:
            fdef.append(f"  {tt}")
        fdef = "\n".join(fdef)
        return fdef

    def gen_fun(self, name, **kwargs):
        """Return a python function that executes a set of tasks in order of dependencies:
        name: name of the functions
        kwards:
            the keys are used to defined the argument name of the functions
            the values are the refs that will be set
        """
        fdef = self.mk_fun(name, **kwargs)
        gbl = {}
        lcl = {}
        gbl.update((k, r._owner) for k, r in self.containers.items())
        exec(fdef, gbl, lcl)
        return lcl[name]

    def plot_deps(self, start=None, **kwargs):
        """Plot a graph of task and target dependencies from start.

        Possible backend:
            mpl: generate a figure in matplotlib
            os: generate a file /tmp/out.png and use `display` to show it
            ipy: use Ipython facility for Jupyter notebooks

        ftype: png, svg, or pdf
        """
        from pydot import Dot, Node, Edge

        if start is None:
            start = list(self.rdeps)
        pdot = Dot("g", graph_type="digraph", rankdir="LR")
        for task in self.find_tasks(start):
            tn = Node(str(task), label=str(task.taskid), shape="oval")
            pdot.add_node(tn)
            for tt in task.targets:
                pdot.add_node(Node(str(tt), shape="box"))
                pdot.add_edge(Edge(tn, str(tt), color="blue"))
            for tt in task.dependencies:
                pdot.add_node(Node(str(tt), shape="box"))
                pdot.add_edge(Edge(str(tt), tn, color="blue"))

        plot_pdot(pdot, **kwargs)
        return pdot

    def plot_tasks(self, start=None, **kwargs):
        """Plot a graph of task dependencies

        Possible backend:
            mpl: generate a figure in matplotlib
            os: generate a file /tmp/out.png and use `display` to show it
            ipy: use Ipython facility for Jupyter notebooks

        ftype: png, svg, or pdf
        """
        from pydot import Dot, Node, Edge

        if start is None:
            start = list(self.rdeps)
        pdot = Dot("g", graph_type="digraph", rankdir="LR")
        for task in self.find_tasks(start):
            tn = Node(str(task.taskid), shape="circle")
            pdot.add_node(tn)
            for dep in task.dependencies:
                for tt in self.tartasks[dep]:
                    pdot.add_edge(Edge(str(tt), tn, color="blue"))

        plot_pdot(pdot, **kwargs)
        return pdot

    def dump(self):
        """Dump in json all ExprTask defined in the manager"""
        data = [
            (str(tt.taskid), str(tt.expr))
            # for t in self.find_tasks(self.rdeps)
            for tt in self.tasks.values()
            if isinstance(tt, ExprTask)
        ]
        return data

    def load(self, dump, dct=None):
        """Reload the expressions in `dump` using container in `dct`

        dump: List[Tuple[MutableRef, ARef]]
            List of pairs representing an expression lhs = rhs for each pair
            (lhs, rhs).
        dct: Optional[Dict[str, Ref]]
            Dictionary of named references of containers, if unspecified
            assume that all containers are in the manager.
        """

        if dct is None:
            dct = self.containers
        for lhs, rhs in dump:
            lhs = eval(lhs, {}, dct)
            rhs = eval(rhs, {}, dct)
            task = ExprTask(lhs, rhs)
            if lhs in self.tasks:
                self.unregister(lhs)
            self.register(task)

    def newenv(self, label="_", data=None):
        """Create a ref with a DepEnv environment in the manager."""
        if data is None:
            data = AttrDict()
        ref = self.ref(data, label=label)
        return DepEnv(data, ref)

    def refattr(self, container=None, label="_"):
        """Create a ref which translates attribute access to item access.

        Useful for accessing globals() like dictionaries.
        """
        if container is None:
            container = AttrDict()
        objref = ObjectAttrRef(container, label, self)
        assert label not in self.containers
        self.containers[label] = objref
        return objref

    def cleanup(self):
        """Remove empty sets from dicts."""
        for dct in self.rdeps, self.rtasks, self.deptasks, self.tartasks:
            for kk, ss in list(dct.items()):
                if len(ss) == 0:
                    del dct[kk]

    def copy(self):
        """Create a copy of in new manager."""
        other = Manager()
        other.containers = deepcopy(self.containers)
        other.tasks = deepcopy(self.tasks)
        other.rdeps = deepcopy(self.rdeps)
        other.rtasks = deepcopy(self.rtasks)
        other.deptasks = deepcopy(self.deptasks)
        other.tartasks = deepcopy(self.tartasks)
        return other

    def clone(self):
        """Regenerate a new manager.

        This differs to copy in that the internal data structures are
        regenerated from the tasks, instead of being copied. This ensures
        the consistency of the new manager at the expense of performance.
        """
        other = Manager()
        other.containers.update(self.containers)
        for task in self.tasks.values():
            other.register(task)
        other.cleanup()
        return other

    def verify(self, dcts=("rdeps", "rtasks", "deptasks", "tartasks")):
        """Verify the consistency of the manager.

        Parameters
        ----------
        dcts: Iterable[str]
            Names of the internal data structures to check.
        """
        self.cleanup()
        other = self.clone()
        for dct in dcts:
            odct = getattr(other, dct)
            sdct = getattr(self, dct)
            for kk, ss in list(sdct.items()):
                if set(ss) != set(odct[kk]):
                    print(f"{dct}[{kk}] not consistent")
                    print(f"{dct}[{kk}] self - check:", set(ss) - set(odct[kk]))
                    print(f"{dct}[{kk}] check - self:", set(odct[kk]) - set(ss))
                    raise (ValueError(f"{self} is not consistent in {dct}[{kk}]"))

    def refresh(self):
        """Regenerate the internal data structures from the tasks."""
        self.rdeps = defaultdict(RefCount)
        self.rtasks = defaultdict(RefCount)
        self.deptasks = defaultdict(RefCount)
        self.tartasks = defaultdict(RefCount)
        for task in self.tasks.values():
            self.register(task)
        self.cleanup()



# copyright ############################### #
# This file is part of the Xdeps Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import xdeps
import math

el = {"el": {"a": 1, "b": 2}}
va = {"a": 2, "b.c": 4}


def test_madxeval():
    ma = xdeps.madxutils.MadxEval(va, math, el)
    assert ma.eval("+1+2^-2") == 1 + 2**-2
    assert ma.eval("b.c") == va["b.c"]
    assert ma.eval("sin(3)^2") == math.sin(3) ** 2
    assert ma.eval("el->a*el->b") == el["el"]["a"] * el["el"]["b"]


def test_madxeval_attr():
    el2 = {"el": type("el", (), el["el"])}
    ma = xdeps.madxutils.MadxEval(va, math, el2, get="attr")
    assert ma.eval("el->a*el->b") == el2["el"].a * el2["el"].b



import os
import pathlib
import re
from typing import Collection
from collections import namedtuple

import numpy as np

gblmath = {"np": np}
for k, fu in np.__dict__.items():
    if type(fu) is np.ufunc:
        gblmath[k] = fu


def _to_str(arr, digits, fixed="g", max_len=None):
    """Covert an array to an array of string representations.

    Parameters
    ----------
    arr : array-like
        Array to convert.
    digits : int
        Number of digits to use for floats.
    fixed : str
        If 'g', use general format (total number of digits is `digits`, if
        necessary use exponent notation). If 'f', use fixed point notation
        (exactly `digits` digits after the decimal point).
    max_len : int or None
        If not None, truncate strings to this length.
    """
    if len(arr.shape) > 1:
        # If array of arrays, just show the shape
        out = np.repeat(f"<array of shape {arr.shape[1:]}>", arr.shape[0])
    elif arr.dtype.kind in "SU":
        # Keep strings
        out = arr
    elif arr.dtype.kind in "iu":
        # Print out integers in full
        out = np.char.mod("%d", arr)
    elif arr.dtype.kind in "f":
        # Format floats.
        # First generate a format string like "%.<digits>g", then use it on
        # each element.
        add = 3 if fixed == "f" else 7
        fmt = "%%%d.%d%s" % (digits + add, digits, fixed)
        out = np.char.mod(fmt, arr)
    elif arr.dtype.kind == "O" and isinstance(arr[0], Collection):
        # If array of collections (array with dtype=object) or list, give shape
        lengths = []
        for entry in arr:
            if isinstance(entry, str):
                lengths.append(entry)
            elif isinstance(entry, np.ndarray):
                lengths.append(f"<array of shape {entry.shape}>")
            elif isinstance(entry, Collection):
                lengths.append(f"<collection of length {len(entry)}>")
            else:
                lengths.append(str(entry))
        out = np.array(lengths)
    else:
        # Any other flat array: stringify.
        out = arr.astype(f"U")

    if max_len is not None:
        old_out = out.copy()
        out = []
        for ii in range(len(old_out)):
            out.append(
                str(
                    old_out[ii][:max_len] + " ..."
                    if len(old_out[ii]) > max_len
                    else old_out[ii]
                )
            )
        out = np.array(out)
    return out


class Mask:
    def __init__(self, table):
        self.table = table

    def __getitem__(self, key):
        """
        t.mask[1] -> row
        t.mask['a'] -> pattern
        l.mask[10:20]-> range
        l.mask['a':'b'] -> name range
        l.mask['a':'b':'myname'] -> name range with 'myname' column
        l.mask[-2:2:'x'] -> name value range with 'x' column
        l.mask[-2:2:'x',...] -> & combinations
        """
        mask = np.zeros(self.table._nrows, dtype=bool)
        if isinstance(key, int):
            mask = np.zeros(self.table._nrows, dtype=bool)
            mask[key] = True
            return mask
        elif isinstance(key, str):
            if self.table._offset_sep in key or self.table._count_sep in key:
                mask[:] = self[key:key]  # use slice (next elif)
            else:
                mask[:] = self.table._get_name_mask(key, self.table._index)
            mask[:] = self.table._get_name_mask(key, self.table._index)
            if self.table._error_on_row_not_found and not mask.any():
                raise IndexError(
                    f"Cannot find `{key}` in table index `{self.table._index}`"
                )
        elif hasattr(key, "dtype"):
            if key.dtype.kind in "SUO":
                # mask[self.table._get_names_indices(key)] = True
                return self.table._get_names_indices(key)  # preserve key order
            else:
                mask[key] = True
        elif isinstance(key, list):
            if len(key) > 0 and isinstance(key[0], str):
                # mask[self.table._get_names_indices(key)] = True
                return self.table._get_names_indices(key)  # preserve key order
            else:
                mask[key] = True
        elif isinstance(key, slice):
            ia = key.start
            ib = key.stop
            ic = key.step
            if isinstance(ia, str) or isinstance(ib, str):
                if ic is None:
                    ic = self.table._index
                if ia is not None:
                    ia = self.table._get_name_indices(ia, ic)[0]
                if ib is not None:
                    ib = self.table._get_name_indices(ib, ic)[-1] + 1
                mask[ia:ib] = True
            elif isinstance(ic, str):
                col = self.table._data[ic]
                if ia is None and ib is None:
                    mask |= True
                elif ia is not None and ib is None:
                    mask |= col <= ib
                elif ib is not None and ia is None:
                    mask |= col >= ia
                else:
                    mask |= (col >= ia) & (col <= ib)
            else:
                mask[ia:ib:ic] = True
        elif isinstance(key, tuple):
            mask = self[key[0]]
            if len(key) > 1:
                mask &= self[key[1:]]

        return mask


class _RowView:
    def __init__(self, table):
        self.table = table

    def __getitem__(self, rows):
        return self.table._get_rows_cols(rows, None)

    def __iter__(self):
        res_type = namedtuple("Row", self.table._col_names)
        for ii in range(len(self.table)):
            yield res_type(*[self.table[cc, ii] for cc in self.table._col_names])

    def _at(self, index, as_dict=False):
        if as_dict:
            return {cc: self.table[cc, index] for cc in self.table._col_names}
        else:
            res_type = namedtuple("Row", self.table._col_names)
            return res_type(*[self.table[cc, index] for cc in self.table._col_names])


class _ColView:
    def __init__(self, table):
        self.table = table

    def __getitem__(self, cols):
        return self.table._get_rows_cols(None, cols, force_table=True)

    @property
    def names(self):
        return self.table._col_names

    def __repr__(self):
        return "<" + " ".join(self.table._col_names) + ">"


class _View:
    def __init__(self, data, index):
        self.data = data
        self.index = index

    def __getitem__(self, k):
        return self.data[k][self.index]

    def __len__(self):
        # k = list(self.data)[0]
        k = self.data._col_names[0]
        return len(self.data[k])

    def get(self, k, default=None):
        if k == "__tracebackhide__":  # to avoid issues in ipython
            return None
        return self.data.get(k, default)[self.index]

    def __repr__(self):
        return f"<{sum(self.index)} rows>"

    def __contains__(self, k):
        return k in self.data

    def __iter__(self):
        return iter(self.data[self.table._index])


class Table:

    _error_on_row_not_found = False

    def __init__(
        self,
        data,
        col_names=None,
        index="name",
        header=None,
        count_sep="##",
        offset_sep="%%",
        index_cache=None,
        cast_strings=True,
        regex_flags=re.IGNORECASE,
    ):
        _data = data.copy()
        _col_names = list(data.keys() if col_names is None else col_names)
        for kk in _col_names:
            vv = data[kk]
            if not hasattr(vv, "dtype"):
                raise ValueError(f"Column `{kk}` is not a numpy array")
            else:
                if cast_strings and vv.dtype.kind in "SU":
                    vv = np.array(vv, dtype=object)
            _data[kk] = vv

        nrows = set(len(_data[cc]) for cc in _col_names)
        assert len(nrows) == 1

        init = {
            "_col_names": _col_names,
            "cols": _ColView(self),
            "_count_sep": count_sep,
            "_data": data.copy(),
            "header": header,
            "_index_cache": index_cache,
            "_index": index,
            "mask": Mask(self),
            "_nrows": nrows.pop(),
            "_offset_sep": offset_sep,
            "_regex_flags": regex_flags,
            "rows": _RowView(self),
        }

        for kk, vv in init.items():
            object.__setattr__(self, kk, vv)

    @classmethod
    def from_pandas(cls, df, index=None, lowercase=False):
        if index is None:
            index = df.index.name
            if index is None and "NAME" in df.columns:
                index = "NAME"
        if lowercase:
            df.columns = df.columns.str.lower()
            index = index.lower()
        col_names = list(df.columns)
        data = {cc: df[cc].values for cc in col_names}
        return cls(data, col_names=col_names, index=index)

    @classmethod
    def from_csv(cls, filename, index=None, col_names=None, **kwargs):
        import pandas as pd

        df = pd.read_csv(filename, **kwargs)
        if index is None and "NAME" in df.columns:
            index = "NAME"

        print(23423, index)
        return cls.from_pandas(df, index=index, col_names=col_names)

    @classmethod
    def from_tfs(cls, filename, index=None, lowercase=True):
        from tfs import read_tfs

        df = read_tfs(filename)
        return cls.from_pandas(df, index=index, lowercase=lowercase)

    def to_pandas(self, index=None, columns=None):

        if columns is None:
            columns = self._col_names

        import pandas as pd

        df = pd.DataFrame(self._data, columns=self._col_names)
        if index is not None:
            df.set_index(index, inplace=True)
        return df

    def _get_index(self):
        if self._index in self._data:
            return self._data[self._index]
        else:
            raise ValueError(f"Cannot find `{self._index}` in table")

    def _get_index_cache(self):
        if self._index_cache is None:
            col = self._get_index()
            dct = {}
            count = {}
            for ii, nn in enumerate(col):
                cc = count.get(nn, -1) + 1
                dct[(nn, cc)] = ii
                count[nn] = cc
            object.__setattr__(self, "_index_cache", dct)
        return self._index_cache

    def _get_row_col_fast(self, row, col, rep=0):
        cache = self._get_index_cache()
        idx = cache[(row, rep)]
        return self[col, idx]

    def _split_name_count_offset(self, name):
        ss = name.split(self._count_sep)
        name = ss[0]
        count = None if len(ss) == 1 else int(ss[1])
        ss = name.split(self._offset_sep)
        name = ss[0]
        offset = 0 if len(ss) == 1 else int(ss[1])
        return name, count, offset

    def _get_name_mask(self, name, col):
        name, count, offset = self._split_name_count_offset(name)
        if col == self._index:
            tryout = self._get_index_cache().get((name, count))
            if tryout is not None:
                mask = np.zeros(self._nrows, dtype=bool)
                mask[tryout] = True
                return mask
        col = self._data[col]
        regex = re.compile(name, re.IGNORECASE)
        it = (regex.fullmatch(rr) is not None for rr in col)
        mask = np.fromiter(it, count=self._nrows, dtype=bool)
        if count is not None:
            idx = np.where(mask)[0][count]
            mask = np.zeros(self._nrows, dtype=bool)
            mask[idx] = True
        return mask

    def _get_name_indices(self, name, col):
        name, count, offset = self._split_name_count_offset(name)
        if col == self._index:
            idx = self._get_index_cache().get((name, count))
            if idx is not None:
                return [idx + offset]
        regex = re.compile(name, self._regex_flags)
        lst = []
        cnt = -1
        for ii, nn in enumerate(self._data[col]):
            if regex.fullmatch(nn):
                cnt += 1
                if count is None or count == cnt:
                    lst.append(ii)
        return np.array(lst, dtype=int) + offset

    def _get_name_index(self, name, col):
        return self._get_name_indices(name, col)[0]

    def _get_names_indices(self, names):
        dct = self._get_index_cache()
        lst = []
        for name in names:
            name, count, offset = self._split_name_count_offset(name)
            if self._error_on_row_not_found and name not in self[self._index]:
                raise IndexError(f"Cannot find `{name}` in table index `{self._index}`")
            if count is None:
                count = 0
            lst.append(dct[(name, count)] + offset)
        return np.array(lst, dtype=int)

    def __getattr__(self, key):
        if key in self._data:
            return self._data[key]
        else:
            raise AttributeError(f"Cannot find `{key}` in table")

    def __len__(self):
        k = self._col_names[0]
        return len(self._data[k])

    def keys(self, exclude_columns=False):
        if exclude_columns:
            return [kk for kk in self._data.keys() if kk not in self._col_names]
        return self._data.keys()

    def values(self):
        return self._data.values()

    def items(self):
        return self._data.items()

    def pop(self, key):
        res = self._data.pop(key)
        if key in self._col_names:
            self._col_names.remove(key)
        return res

    def __dir__(self):
        return super().__dir__() + list(self._data.keys())

    def __iter__(self):
        return self._data.__iter__()

    def __contains__(self, key):
        return self._data.__contains__(key)

    def __setitem__(self, key, val):
        if key == self._index:
            object.__setattr__(self, "_index_cache", None)
        if key in self._col_names:
            self._data[key][:] = val
        else:
            self._data[key] = val
            if hasattr(val, "__iter__") and len(val) == self._nrows:
                self._col_names.append(key)

    def __delitem__(self, key, val):
        if key in self._col_names:
            self._col_names.remove(key)
        del self._data[key]

    __setattr__ = __setitem__

    def __repr__(self):
        n = self._nrows
        c = len(self._col_names)
        ns = "s" if n != 1 else ""
        cs = "s" if c != 1 else ""
        out = [f"{self.__class__.__name__}: {n} row{ns}, {c} col{cs}"]
        if self._nrows < 10000:
            show = self.show(output=str)
            out.append(show)
        return "\n".join(out)

    def __getitem__(self, args):
        if type(args) is str and args in self._data:
            return self._data[args]
        if type(args) is tuple:  # multiple args
            if len(args) == 0:
                cols = None
                rows = None
            elif len(args) == 1:
                cols = args[0]
                rows = None
            elif len(args) == 2:
                cols = args[0]
                rows = args[1]
                if isinstance(rows, str) and isinstance(cols, str):
                    return self._get_row_col_fast(rows, cols)
            else:
                cols = args[0]
                rows = args[1:]
        else:  # one arg
            cols = args
            rows = None
        return self._get_rows_cols(rows, cols)

    def _get_view_col_list(self, rows, cols):
        # select rows
        if rows is None:
            view = self._data
        else:
            row_index = self.mask[rows]
            view = _View(self._data, row_index)

        # select cols
        if cols is None or cols == slice(None, None, None):
            # Sort columns to put array fields at the end
            col_list = sorted(
                self._col_names,
                key=lambda x: len(self._data[x].shape),
            )
        elif type(cols) is str:
            col_list = cols.split()
        else:
            col_list = list(cols)

        return view, col_list

    def _get_rows_cols(self, rows, cols, force_table=False):
        view, col_list = self._get_view_col_list(rows, cols)

        # return data
        if len(col_list) == 1 and not force_table:
            try:
                cc = eval(col_list[0], gblmath, view)
            except NameError:
                raise KeyError(
                    f"Column `{col_list[0]}` could not be found or "
                    "is not a valid expression"
                )
            if len(cc) == 1:
                return cc[0]  # scalar
            else:
                return cc  # array
        else:
            if self._index not in col_list:
                col_list.insert(0, self._index)
            data = {}
            for cc in col_list:
                if cc in view:
                    data[cc] = view[cc]
                else:
                    try:
                        data[cc] = eval(cc, gblmath, view)
                    except NameError as ex:
                        print(ex)
                        raise KeyError(
                            f"Column or expr `{cc}` could not be found or "
                            "is not a valid expression"
                        )
            for kk in self.keys(exclude_columns=True):
                data[kk] = self._data[kk]
            return self.__class__(
                data, index=self._index, count_sep=self._count_sep, col_names=col_list
            )  # table

    def show(
        self,
        rows=None,
        cols=None,
        maxrows=20,
        maxwidth="auto",
        output=None,
        digits=6,
        fixed="g",
        header=True,
        max_col_width=None,
    ):
        view, col_list = self._get_view_col_list(rows, cols)

        # index always first
        if self._index in col_list:
            col_list.remove(self._index)
        col_list.insert(0, self._index)

        cut = -1
        viewrows = len(view)
        if maxrows is not None and output is None and viewrows > maxrows:
            cut = maxrows // 2

        if maxwidth == "auto":
            try:
                maxwidth = os.get_terminal_size().columns - 5
                if maxwidth < 10 or maxwidth > 10000:
                    raise ValueError("Terminal width too big or too small.")
            except (OSError, ValueError):
                maxwidth = 100

        data = []
        width = 0
        fmt = []
        header_line = []
        for cc in col_list:
            if cc in view:
                coldata = view[cc]
            else:
                coldata = eval(cc, gblmath, view)
            if cut > 0:
                coldata = np.r_[coldata[:cut], coldata[cut:]]
            coltype = coldata.dtype.kind
            col = _to_str(coldata, digits, fixed, max_len=max_col_width)
            colwidth = int(col.dtype.str[2:])
            if len(cc) > colwidth:
                colwidth = len(cc)
            width += colwidth + 1
            if width < maxwidth:
                if coltype in "SUO":
                    fmt.append("%%-%ds" % (colwidth))
                else:
                    fmt.append("%%%ds" % colwidth)
                header_line.append(fmt[-1] % str(cc))
                data.append(col)
            else:
                header_line.append("...")
                break

        result = []
        if header:
            result.append(" ".join(header_line))
        for ii in range(len(col)):
            row = " ".join([ff % col[ii] for ff, col in zip(fmt, data)])
            result.append(row)
            if ii == cut:
                result.append("...")
        result = "\n".join(result)
        if output is None:
            print(result)
        elif output is str:
            return result
        elif hasattr(output, "write"):
            output.write(result)
        else:
            output = pathlib.Path(output)
            with open(output, "w") as fh:
                fh.write(result)

    @property
    def _t(self):
        data = {"columns": np.array(self._col_names)}
        for nn in range(len(self)):
            data[f"row{nn}"] = np.array([str(self[cc][nn]) for cc in self._col_names])
        return Table(data, index="columns", col_names=list(data.keys()))

    def _update(self, data):
        if hasattr(self, "_data"):
            data = self._data
        for name, value in data.items():
            if hasattr(value, "__len__") and len(value) == len(self):
                if name not in self._data:
                    self._col_names.append(name)
            self._data[name] = value

    def _append_row(self, row):
        for col in self._col_names:
            self._data[col] = np.r_[self._data[col], [row[col]]]
